from __future__ import division
from __future__ import print_function
from numpy.core.arrayprint import dtype_is_implied
from tigre.utilities.parkerweight import parkerweight
import numpy as np
from scipy.fft  import fft, ifft

import warnings

import numpy as np
from tigre.utilities.parkerweight import parkerweight


# TODO: Fix parker
def filtering(proj, geo, angles, parker, verbose=False):
    if parker:
        proj=parkerweight(proj.transpose(0,2,1),geo,angles,parker).transpose(0,2,1)

    filt_len=max(64,2**nextpow2(2*max(geo.nDetector)))
    ramp_kernel=ramp_flat(filt_len)

    d=1
    filt=filter(geo.filter,ramp_kernel[0],filt_len,d,verbose=verbose)
    filt=np.kron(np.ones((np.int64(geo.nDetector[0]),1),dtype=np.float32),filt)

    padding = int((filt_len-geo.nDetector[1])//2 )
    scale_factor = (geo.DSD[0]/geo.DSO[0]) * (2 * np.pi/ len(angles)) / ( 4 * geo.dDetector[0] ) 

    #filter 2 projection at a time packing in to complex container
    fproj=np.empty((geo.nDetector[0],filt_len),dtype=np.complex64)
    for i in range(0,angles.shape[0]-1,2):
        fproj.fill(0)
        fproj.real[:,padding:padding+geo.nDetector[1]]=proj[i]
        fproj.imag[:,padding:padding+geo.nDetector[1]]=proj[i+1]

        fproj=fft(fproj,axis=1)
        fproj=fproj*filt
        fproj=ifft(fproj,axis=1)

        proj[i]=fproj.real[:,padding:padding+geo.nDetector[1]] * scale_factor
        proj[i+1]=fproj.imag[:,padding:padding+geo.nDetector[1]] * scale_factor

    #if odd number of projections filter last solo
    if angles.shape[0] % 2:
        fproj.fill(0)
        fproj.real[:,padding:padding+geo.nDetector[1]]=proj[angles.shape[0]-1]

        fproj=fft(fproj,axis=1)
        fproj=fproj*filt
        fproj=np.real(ifft(fproj,axis=1))     
        proj[angles.shape[0]-1]=fproj[:,padding:padding+geo.nDetector[1]] * scale_factor

    return proj


def ramp_flat(n, verbose=False):
    nn = np.arange(-n / 2, n / 2)
    h = np.zeros(nn.shape, dtype=np.float32)
    h[int(n / 2)] = 1 / 4
    odd = nn % 2 == 1
    h[odd] = -1 / (np.pi * nn[odd]) ** 2
    return h, nn


def filter(filter, kernel, order, d, verbose=False):
    f_kernel = abs(np.fft.fft(kernel)) * 2

    filt = f_kernel[: int((order / 2) + 1)]
    w = 2 * np.pi * np.arange(len(filt)) / order

    if filter in {"ram_lak", None}:
        if filter is None and verbose:
            warnings.warn("no filter selected, using default ram_lak")
    elif filter == "shepp_logan":
        filt[1:] *= np.sin(w[1:] / (2 * d)) / (w[1:] / (2 * d))
    elif filter == "cosine":
        filt[1:] *= np.cos(w[1:] / (2 * d))
    elif filter == "hamming":
        filt[1:] *= 0.54 + 0.46 * np.cos(w[1:] / d)
    elif filter == "hann":
        filt[1:] *= (1 + np.cos(w[1:]) / d) / 2
    else:
        raise ValueError("filter not recognised: " + str(filter))

    filt[w > np.pi * d] = 0
    filt = np.hstack((filt, filt[1:-1][::-1]))
    return filt.astype(np.float32)


def nextpow2(n):
    i = 1
    while (2 ** i) < n:
        i += 1
    return i



from __future__ import division
from __future__ import print_function

import numpy as np
import tigre


def svd_power_method(arr, geo, angles, **kwargs):
    """
    method to determine the largest singular value for the rectangular A
    projection matrix.

    :param arr: (np.ndarray, dtype=np.float32)
        random vector
    :param geo: (tigre.geometry())
        geometry of the set up
    :param angles: (np.ndarray, dtype=np.float32)
        angles of projection
    :param kwargs:
        optional arguments
    :keyword maxiter: (int)
        number of iterations to run unless margin of error epsilon has
        been reached. Default: 100
    :keyword epsilon: (float)
        margin of error for the algorithm. Default: 0.001
    :keyword verbose: (bool)
        print stuff. Default: False

    :return: (float)
        the largest singular value of the A matrix.
    """
    maxiter = 100
    if "maxiter" in kwargs:
        maxiter = kwargs["maxiter"]
    # very optimistic value for epsilon (it turns out)
    epsilon = 0.001
    if "epsilon" in kwargs:
        epsilon = kwargs["epsilon"]
    verbose = False
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]

    error = np.inf

    arr_old = arr.copy()
    i = 0
    while error >= epsilon:
        if verbose:
            if i == 0:
                print("using SVD power method to calculate " "Lipschitz const")
            if np.mod(i, 10) == 0:
                print("error for val is:" + str(error))
        # very verbose and inefficient for now
        omega = tigre.Ax(arr_old, geo, angles)
        alpha = np.linalg.norm(omega)
        u = (1.0 / alpha) * omega
        z = tigre.Atb(u, geo, angles)
        beta = np.linalg.norm(z)
        arr = (1.0 / beta) * z
        error = np.linalg.norm(tigre.Ax(arr, geo, angles) - beta * u)
        sigma = beta
        arr_old = arr
        i += 1
        if i >= maxiter:
            return sigma

    return sigma


# http://www.anstuocmath.ro/mathematics/anale2015vol2/Bentbib_A.H.__Kanber_A..pdf



from __future__ import division

import copy

import numpy as np
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.algorithms.single_pass_algorithms import FDK
from tigre.utilities.Ax import Ax
from tigre.utilities.im3Dnorm import im3DNORM




class initASD_POCS(IterativeReconAlg):  
    """
    ASD_POCS solves the ASD_POCS total variation constrained image in 3D
    tomography

    Parameters
    ----------
    :param proj: (np.ndarray, dtype=np.float32)
    Input data, shape = (geo.nDector, nangles)

    :param geo: (tigre.geometry)
    Geometry of detector and image (see examples/Demo code)

    :param angles: (np.ndarray , dtype=np.float32)
    angles of projection, shape = (nangles,3)

    :param niter: (int)
    number of iterations for reconstruction algorithm

    :param kwargs: (dict)
    optional parameters

    Keyword Arguments
    -----------------
    :keyword blocksize: (int)
        number of angles to be included in each iteration
        of proj and backproj in OS_SART update

    :keyword lmbda: (np.float64)
        Sets the value of the hyperparameter for SART iteration.

    :keyword lmbda_red: (np.float64)
        Reduction of lmbda every iteration
        lmbda=lmbda_red*lmbda. Default is 0.99

    :keyword init: (str)
        Describes different initialization techniques.
            None      : Initializes the image to zeros (default)
            "FDK"      : intializes image to FDK reconstrucition

    :keyword verbose:  (Boolean)
        Feedback print statements for algorithm progress
        default=True

    :keyword OrderStrategy : (str)
        Chooses the subset ordering strategy. Options are:
            "ordered"        : uses them in the input order, but
                            divided
            "random"         : orders them randomply
            
    :keyword tviter: (int)
        Number of iterations of minmizeTV to be performed for each iter-
        ation. Default: 20

    :keyword alpha: (float)
        Defines the TV hyperparameter.
        Default: 0.002

    :keyword alpha_red: (float)
        Defines the reduction rate of the hyperparameter
        Default: 0.95

    :keyword rmax: (float)
        The maximum allowed image/TV update ration. If the TV
        update changes the image more than this, the parameter
        will be reduced. Default: 0.95

    :keyword maxl2err: (float)
        Maximum L2 error to accept an image as valid. This
        parameter is crucial for the algorithm, determines at
        what point an image should not be updated further.
        Default is 20% of the FDK L2 norm.
    --------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
    --------------------------------------------------------------------
    Coded by:          MATLAB (original code): Ander Biguri
                        PYTHON : Reuben Lindroos
    """

    def __init__(self, proj, geo, angles, niter, **kwargs):

        
        # kwargs.update(dict(regularisation="minimizeTV"))
        # if "blocksize" not in kwargs:
        #     kwargs.update(dict(blocksize=1))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)       
        self.alpha = 0.002 if "alpha" not in kwargs else kwargs["alpha"]
        self.alpha_red = 0.95 if "alpha_red" not in kwargs else kwargs["alpha_red"]
        self.rmax = 0.95 if "rmax" not in kwargs else kwargs["rmax"]
        if "maxl2err" not in kwargs:
            self.epsilon = (
                im3DNORM(Ax(FDK(proj, geo, angles, gpuids=self.gpuids), geo, angles) - proj, 2)
                * 0.2
            )
        else:
            self.epsilon = kwargs["maxl2err"]
        self.numiter_tv = 20 if "tviter" not in kwargs else kwargs["tviter"]
        self.beta = self.lmbda
        self.beta_red = self.lmbda_red
        

    # Override
    def run_main_iter(self):
        stop_criteria = False
        n_iter = 0
        while not stop_criteria:
            if self.verbose:
                self._estimate_time_until_completion(n_iter)

            res_prev = copy.deepcopy(self.res)
            getattr(self, self.dataminimizing)()
            if self.Quameasopts is not None:
                self.error_measurement(res_prev, n_iter)
            n_iter += 1
            g = Ax(self.res, self.geo, self.angles, gpuids=self.gpuids)
            dd = im3DNORM(g - self.proj, 2)
            dp_vec = self.res - res_prev
            dp = im3DNORM(dp_vec, 2)

            if n_iter == 1:
                dtvg = self.alpha * dp

            res_prev = copy.deepcopy(self.res)
            self.res = getattr(self, self.regularisation)(self.res, dtvg)
            dg_vec = self.res - res_prev
            dg = im3DNORM(dg_vec, 2)

            if dg > self.rmax * dp and dd > self.epsilon:
                dtvg = dtvg * self.alpha_red

            self.beta *= self.beta_red
            c = np.dot(dg_vec.reshape(-1,), dp_vec.reshape(-1,)) / max(
                dg * dp, 1e-6
            )  # reshape ensures no copy is made.
            if (c < -0.99 and dd <= self.epsilon) or self.beta < 0.005 or n_iter >= self.niter:
                if self.verbose:
                    print(
                        "\n"
                        "     Stop criteria met: \n"
                        "     c = " + str(c) + "\n"
                        "     beta = " + str(self.beta) + "\n"
                        "     iter = " + str(n_iter) + "\n"
                    )
                stop_criteria = True


class ASD_POCS(initASD_POCS):
    __doc__ = initASD_POCS.__doc__
    
    def __init__(self, proj, geo, angles, niter, **kwargs):
        
        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        kwargs.update(dict(regularisation="minimizeTV"))
        
        initASD_POCS.__init__(self, proj, geo, angles, niter, **kwargs)
       

asd_pocs = decorator(ASD_POCS, name="asd_pocs")


class AwASD_POCS(initASD_POCS):  
    __doc__ = ("    AwASD_POCS is the Adaptive Weighted TV (edge preserving) version of ASD_POCS\n\n"
               "        :extra kwargs delta: (float)\n"
               "            Control amount of smoothing at edges of the image\n"
               "            Default delta = -0.005\n") + initASD_POCS.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        kwargs.update(dict(regularisation="minimizeAwTV"))
        self.delta = np.float32(-0.005) if "delta" not in kwargs else kwargs["delta"] 
        
        initASD_POCS.__init__(self, proj, geo, angles, niter, **kwargs)


awasd_pocs = decorator(AwASD_POCS, name="awasd_pocs")


class OS_ASD_POCS(initASD_POCS):
    __doc__ = ("    Oriented Subsets version of ASD_POCS \n\n"
               "    Default blocksize = 20\n") + initASD_POCS.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.blocksize = 20 if "blocksize" not in kwargs else kwargs["blocksize"]
        kwargs.update(dict(regularisation="minimizeTV"))
        
        initASD_POCS.__init__(self, proj, geo, angles, niter, **kwargs)


os_asd_pocs = decorator(OS_ASD_POCS, name="os_asd_pocs")


class OS_AwASD_POCS(initASD_POCS): 
    __doc__ = (
        "    Oriented Subsets and Adaptive Weighted TV version of AwASD_POCS\n\n"
        "    Default blocksize = 20\n\n"
        "        :extra kwargs delta: (float)\n"
        "            Control amount of smoothing at edges of the image\n"
        "            Default delta = -0.005\n"
        ) + initASD_POCS.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.blocksize = 20 if "blocksize" not in kwargs else kwargs["blocksize"]
        kwargs.update(dict(regularisation="minimizeAwTV"))
        self.delta = np.float32(-0.005) if "delta" not in kwargs else kwargs["delta"]
        
        initASD_POCS.__init__(self, proj, geo, angles, niter, **kwargs) 


os_awasd_pocs = decorator(OS_AwASD_POCS, name="os_awasd_pocs")




class initPCSD(IterativeReconAlg):
    """
    PCSD solves the reconstruction problem using projection-controlled 
    steepest descent method

    Parameters
    ----------
    :param proj: (np.ndarray, dtype=np.float32)
    Input data, shape = (geo.nDector, nangles)

    :param geo: (tigre.geometry)
    Geometry of detector and image (see examples/Demo code)

    :param angles: (np.ndarray , dtype=np.float32)
    angles of projection, shape = (nangles,3)

    :param niter: (int)
    number of iterations for reconstruction algorithm

    :param kwargs: (dict)
    optional parameters

    Keyword Arguments
    -----------------
    :keyword blocksize: (int)
        number of angles to be included in each iteration
        of proj and backproj in OS_SART update
        
    :keyword lambda: (np.float64)
        Sets the value of the hyperparameter for SART iteration.

    :keyword lambda_red: (np.float64)
        Reduction of lambda every iteration
        lambda=lambda_red*lambda. Default is 0.99

    :keyword init: (str)
        Describes different initialization techniques.
               None      : Initializes the image to zeros (default)
              "FDK"      : intializes image to FDK reconstrucition

    :keyword verbose:  (Boolean)
        Feedback print statements for algorithm progress
        default=True

    :keyword OrderStrategy : (str)
        Chooses the subset ordering strategy. Options are:
                 "ordered"        : uses them in the input order, but
                                    divided
                 "random"         : orders them randomply
                 
    :keyword tviter: (int)
        Number of iterations of minmizeTV to be performed for each iter-
        ation. Default: 20

    :keyword maxl2err: (float)
         Maximum L2 error to accept an image as valid. This
         parameter is crucial for the algorithm, determines at
         what point an image should not be updated further.
         Default is 20% of the FDK L2 norm.
--------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
    --------------------------------------------------------------------
    Coded by:          MATLAB (original code): Ander Biguri
                       PYTHON : Reuben Lindroos



    """

    def __init__(self, proj, geo, angles, niter, **kwargs):
        
        # if "blocksize" not in kwargs:
        #     kwargs.update(dict(blocksize=1))
        #kwargs.update(dict(regularisation="minimizeTV"))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
        if "maxl2err" not in kwargs:
            self.epsilon = (
                im3DNORM(Ax(FDK(proj, geo, angles, gpuids=self.gpuids), geo, angles) - proj, 2)
                * 0.2
            )
        else:
            self.epsilon = kwargs["maxl2err"]
        self.numiter_tv = 20 if "tviter" not in kwargs else kwargs["tviter"]
        self.beta = self.lmbda
        self.beta_red = self.lmbda_red

    # Override
    def run_main_iter(self):
        stop_criteria = False
        n_iter = 0
        d_p_1st = 1
        while not stop_criteria:
            if self.verbose:
                self._estimate_time_until_completion(n_iter)
                
            res_prev = copy.deepcopy(self.res)
            n_iter += 1
            est_proj = Ax(self.res, self.geo, self.angles)
            d_p = im3DNORM(est_proj - self.proj, 2)
            if d_p**2 > self.epsilon:
                getattr(self, self.dataminimizing)()
            dd = im3DNORM(Ax(self.res,self.geo,self.angles)-self.proj, 2)
            dp_vec = self.res - res_prev
            dp = im3DNORM(dp_vec, 2)
            
            dtvg = 1 if n_iter == 1 else d_p / d_p_1st
                
            res_prev = copy.deepcopy(self.res)
            self.res = getattr(self, self.regularisation)(self.res, dtvg)
            dg_vec = self.res - res_prev
            dg = im3DNORM(dg_vec, 2)
            
            if n_iter == 1:
                d_p_1st = im3DNORM(Ax(res_prev,self.geo,self.angles)-self.proj, 2)

            self.beta *= self.beta_red
            c = np.dot(dg_vec.reshape(-1,), dp_vec.reshape(-1,)) / max(dg * dp, 1e-6) # reshape ensures no copy is made. 
            if (c < -0.99 and dd <=
                    self.epsilon) or self.beta < 0.005 or n_iter > self.niter:
                if self.verbose:
                    print("\n"
                          "     Stop criteria met: \n"
                          "     c = " + str(c) + "\n"
                          "     beta = " + str(self.beta) + "\n"
                          "     iter = " + str(n_iter) + "\n")
                stop_criteria = True


class PCSD(initPCSD):
    __doc__ = initPCSD.__doc__
    
    def __init__(self, proj, geo, angles, niter, **kwargs):

        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        kwargs.update(dict(regularisation="minimizeTV"))
        
        initPCSD.__init__(self, proj, geo, angles, niter, **kwargs)
        
    
pcsd = decorator(PCSD, name='pcsd')


class AwPCSD(initPCSD): 
    __doc__ = (        
        "    Adaptive Weighted TV (edge preserving) version of PCSD\n\n"
        "    :extra keyword delta: (float)\n"
        "        Controls amount of smoothing at edges of the image \n"
        "        Default -0.005\n"
        ) + initPCSD.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        kwargs.update(dict(regularisation="minimizeAwTV"))
        self.delta = np.float32(-0.005) if "delta" not in kwargs else kwargs["delta"]
            
        initPCSD.__init__(self, proj, geo, angles, niter, **kwargs)
        

aw_pcsd = decorator(AwPCSD, name="aw_pcsd")


class OS_PCSD(initPCSD):
    __doc__ = (
        "    Oriented Subsets version of PCSD\n\n"
        "        Default blocksize = 20"
        ) + initPCSD.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.blocksize = 20 if "blocksize" not in kwargs else kwargs["blocksize"]
        kwargs.update(dict(regularisation="minimizeTV"))
            
        initPCSD.__init__(self, proj, geo, angles, niter, **kwargs)
        

os_pcsd = decorator(OS_PCSD, name="os_pcsd")


class OS_Aw_PCSD(initPCSD): 
    __doc__ = (
        "    Oriented Subsets and Adaptive Weighted TV version of PCSD\n\n"
        "    Default blocksize = 20\n\n"
        "    :extra keyword delta: (float)\n"
        "        Controls amount of smoothing at edges of the image \n"
        "        Default -0.005\n"
        ) + initPCSD.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.blocksize = 20 if "blocksize" not in kwargs else kwargs["blocksize"]
        kwargs.update(dict(regularisation="minimizeAwTV"))
        self.delta = np.float32(-0.005) if "delta" not in kwargs else kwargs["delta"]
            
        initPCSD.__init__(self, proj, geo, angles, niter, **kwargs)
        
        
os_aw_pcsd = decorator(OS_Aw_PCSD, name="os_aw_pcsd")






from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

import numpy as np
from tigre.utilities.Atb import Atb
from tigre.utilities.filtering import filtering



def FDK(proj, geo, angles, **kwargs):
    """
    solves CT image reconstruction.

    :param proj: np.array(dtype=float32),
    Data input in the form of 3d

    :param geo: tigre.utilities.geometry.Geometry
    Geometry of detector and image (see examples/Demo code)

    :param angles: np.array(dtype=float32)
    Angles of projection, shape = (nangles,3) or (nangles,)

    :param filter: str
    Type of filter used for backprojection
    opts: "shep_logan"
          "cosine"
          "hamming"
          "hann"

    :param verbose: bool
    Feedback print statements for algorithm progress

    :param kwargs: dict
    keyword arguments

    :return: np.array(dtype=float32)

    Usage:
    -------
    >>> import tigre
    >>> import tigre.algorithms as algs
    >>> import numpy
    >>> from tigre.demos.Test_data import data_loader
    >>> geo = tigre.geometry(mode='cone',default_geo=True,
    >>>                         nVoxel=np.array([64,64,64]))
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> src_img = data_loader.load_head_phantom(geo.nVoxel)
    >>> proj = tigre.Ax(src_img,geo,angles)
    >>> output = algs.FDK(proj,geo,angles)

    tigre.demos.run() to launch ipython notebook file with examples.

    --------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
    --------------------------------------------------------------------
    Coded by:           MATLAB (original code): Ander Biguri
                        PYTHON : Reuben Lindroos
    """
    verbose = kwargs["verbose"] if "verbose" in kwargs else False

    gpuids = kwargs["gpuids"] if "gpuids" in kwargs else None
    geo = copy.deepcopy(geo)
    geo.check_geo(angles)
    geo.checknans()
    geo.filter = kwargs["filter"] if "filter" in kwargs else None
    # Weight
    proj_filt = np.zeros(proj.shape, dtype=np.float32)
    xv = np.arange((-geo.nDetector[1] / 2) + 0.5,
                       1 + (geo.nDetector[1] / 2) - 0.5) * geo.dDetector[1]
    yv = np.arange((-geo.nDetector[0] / 2) + 0.5,
                       1 + (geo.nDetector[0] / 2) - 0.5) * geo.dDetector[0]
    (yy, xx) = np.meshgrid(xv, yv)

    w = geo.DSD[0] / np.sqrt((geo.DSD[0] ** 2 + xx ** 2 + yy ** 2))
    np.multiply(proj, w, out=proj_filt)

    proj_filt = filtering(proj_filt, geo, angles, parker=False, verbose=verbose)
    
    return Atb(proj_filt, geo, geo.angles, "FDK", gpuids=gpuids)


fdk = FDK


def fbp(proj, geo, angles, **kwargs):  # noqa: D103
    __doc__ = FDK.__doc__  # noqa: F841
    if geo.mode != "parallel":
        raise ValueError("Only use FBP for parallel beam. Check geo.mode.")
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    verbose = kwargs["verbose"] if "verbose" in kwargs else False
    gpuids = kwargs["gpuids"] if "gpuids" in kwargs else None
    proj_filt = filtering(copy.deepcopy(proj), geox, angles, parker=False, verbose=verbose)
    return Atb(proj_filt, geo, angles, gpuids=gpuids) * geo.DSO / geo.DSD



from __future__ import division
import numpy as np


def _get_angularDistance_index(angles):
    # Sort values according to size
    angles_alpha = angles[:, 0]  # omit theta and psi
    index_alpha = np.argsort(angles_alpha)
    # Save the first values for index_alpha and angles in the new lists
    # (this needs to be done or argmin will return error otherwise)
    new_index = [index_alpha[0]]
    new_angles_alpha = [angles_alpha[index_alpha[0]]]
    # remove the first values from the original arrays
    angles_alpha = np.delete(angles_alpha[index_alpha], 0)
    index_alpha = np.delete(index_alpha, 0)
    # Run through list
    for i in range(len(angles_alpha)):
        # Run the comparison
        compangle = new_angles_alpha[i]
        angles_min = np.argmin(abs(abs(angles_alpha - compangle) - (np.pi / 2)))
        # Save the results to the new lists
        new_angles_alpha.append(angles_alpha[angles_min])
        new_index.append(index_alpha[angles_min])
        # delete old values from the original arrays
        angles_alpha = np.delete(angles_alpha, angles_min)
        index_alpha = np.delete(index_alpha, angles_min)

    return np.array(new_index)


def order_subsets(angles, blocksize, mode):
    """
    Returns
        block_alpha, index_alpha: lists of numpy.ndarray
    """

    # Parse no blocks
    if blocksize is None or blocksize == 1:
        index_alpha = np.arange(angles.shape[0])

        if mode is None or mode == "ordered":
            return list(angles), list(index_alpha)

        if mode == "random" or mode == "random2":
            np.random.shuffle(index_alpha)
            return list(angles[index_alpha]), list(index_alpha)
        if mode == "angularDistance":
            new_index = _get_angularDistance_index(angles)
            new_angles = angles[new_index]
            return list(np.array(new_angles, dtype=np.float32)), list(new_index)
        else:
            raise NameError("mode string not recognised: " + mode)

    # Parse with blocks
    elif blocksize > 1:
        nangles = angles.shape[0]
        block_count = (nangles+blocksize-1)//blocksize
        if mode == "random2":
            new_order = np.arange(nangles)
            np.random.shuffle(new_order)
            index_alpha = [new_order[i*blocksize:(i+1)*blocksize] for i in range(block_count-1)]
            index_alpha.append(new_order[(block_count-1)*blocksize:nangles])
            new_alpha = angles[new_order]
            block_alpha = [new_alpha[i*blocksize:(i+1)*blocksize] for i in range(block_count-1)]
            block_alpha.append(new_alpha[(block_count-1)*blocksize:nangles])
            return block_alpha, index_alpha

        if mode == "angularDistance":
            new_order = _get_angularDistance_index(angles)

            index_alpha = [new_order[i*blocksize:(i+1)*blocksize] for i in range(block_count-1)]
            index_alpha.append(new_order[(block_count-1)*blocksize:nangles])
            new_alpha = angles[new_order]
            block_alpha = [new_alpha[i*blocksize:(i+1)*blocksize] for i in range(block_count-1)]
            block_alpha.append(new_alpha[(block_count-1)*blocksize:nangles])
            return block_alpha, index_alpha

        # using list comprehension to form the blocks.
        oldindex = np.arange(nangles,dtype=np.int32)
        index_alpha = [oldindex[i*blocksize:(i+1)*blocksize] for i in range(0,block_count-1)]
        index_alpha.append(oldindex[(block_count-1)*blocksize:nangles])
        block_alpha = [angles[i*blocksize:(i+1)*blocksize] for i in range(0,block_count-1)]
        block_alpha.append(angles[(block_count-1)*blocksize:nangles])

        if mode is None or mode == "ordered":
            return block_alpha, index_alpha

        if mode == "random" or mode is None:
            new_order = np.arange(len(index_alpha))
            np.random.shuffle(new_order)

            index_alpha = [index_alpha[i] for i in new_order]
            block_alpha = [block_alpha[i] for i in new_order]
            return block_alpha, index_alpha
        else:
            raise NameError("mode string not recognised: " + mode)



cimport numpy as np
import numpy as np
from tigre.utilities.cuda_interface._types cimport Geometry as c_Geometry, convert_to_c_geometry, free_c_geometry
from tigre.utilities.cuda_interface._gpuUtils cimport GpuIds as c_GpuIds, convert_to_c_gpuids, free_c_gpuids


# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

from libc.stdlib cimport malloc, free

cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)
    void PyArray_CLEARFLAGS(np.ndarray arr, int flags)


cdef extern from "Siddon_projection.hpp":
    cdef int siddon_ray_projection(float* img, c_Geometry geo, float** result, float* alphas, int nalpha, const c_GpuIds& gpuids)
cdef extern from "Siddon_projection_parallel.hpp":
    cdef int siddon_ray_projection_parallel(float* img, c_Geometry geo, float** result, float* alphas, int nalpha, const c_GpuIds& gpuids)
cdef extern from "ray_interpolated_projection.hpp":
    cdef int interpolation_projection(float* img, c_Geometry geo, float** result, float* alphas, int nalpha, const c_GpuIds& gpuids)
cdef extern from "ray_interpolated_projection_parallel.hpp":
    cdef int interpolation_projection_parallel(float* img, c_Geometry geo, float** result, float* alphas, int nalpha, const c_GpuIds& gpuids)

def cuda_raise_errors(error_code):
    if error_code:
        raise ValueError('TIGRE: Call to Ax failed')

def _Ax_ext(np.ndarray[np.float32_t, ndim=3] img, geometry, np.ndarray[np.float32_t, ndim=2] angles, projection_type="Siddon", mode="cone", gpuids=None):
    
    cdef c_GpuIds* c_gpuids = convert_to_c_gpuids(gpuids)
    if not c_gpuids:
        raise MemoryError("Error reading GPU Ids")
    
    cdef int total_projections = angles.shape[0]
    cdef unsigned long nDetectorPixels = <unsigned long>geometry.nDetector[0] * <int>geometry.nDetector[1]

    cdef c_Geometry* c_geometry = convert_to_c_geometry(geometry, total_projections)
    if not c_geometry:
        raise MemoryError("Error reading the geometry")
    cdef float* c_projectionsNonPinned = <float*>malloc(nDetectorPixels * total_projections* sizeof(float))
    if not c_projectionsNonPinned:
        raise MemoryError("Error creating memory for projections")
    cdef float** c_projections = <float**> malloc(total_projections * sizeof(float*))
    if not c_projections:
        raise MemoryError("Error creating memory to host projections")
    cdef int i = 0
    for i in range(total_projections):
        c_projections[i] = <float*> &(c_projectionsNonPinned[nDetectorPixels * i])

    angles = np.ascontiguousarray(angles)

    cdef float* c_angles = <float*> angles.data
    if not c_angles:
        raise MemoryError("Error loading angle data")
    if projection_type == "Siddon" or projection_type == "ray-voxel":
        interpolated = False
    elif projection_type == "interpolated":
        interpolated = True
    else:
        print("Warning: Unknown projection_type, using default Siddon")
        interpolated = False

    if mode == "parallel":
        cone_beam = False
    elif mode == "cone":
        cone_beam = True
    else:
        print("Warning: Unknown mode, using default cone beam")
        cone_beam = True

    img = np.ascontiguousarray(img)

    cdef float* c_img = <float*> img.data
    if cone_beam:
        if not interpolated:
            cuda_raise_errors(siddon_ray_projection(c_img, c_geometry[0], c_projections, c_angles, total_projections, c_gpuids[0]))
        else:
            cuda_raise_errors(interpolation_projection(c_img, c_geometry[0], c_projections, c_angles, total_projections, c_gpuids[0]))
    else:
        if not interpolated:
            cuda_raise_errors(siddon_ray_projection_parallel(c_img, c_geometry[0], c_projections, c_angles, total_projections, c_gpuids[0]))
        else:
            cuda_raise_errors(interpolation_projection_parallel(c_img, c_geometry[0], c_projections, c_angles, total_projections, c_gpuids[0]))

 
    cdef np.npy_intp shape[3]
    shape[0] = <np.npy_intp> total_projections
    shape[1] = <np.npy_intp> geometry.nDetector[0]
    shape[2] = <np.npy_intp> geometry.nDetector[1]
    projections = np.PyArray_SimpleNewFromData(3, shape, np.NPY_FLOAT32, c_projectionsNonPinned)
    PyArray_ENABLEFLAGS(projections, np.NPY_OWNDATA)  # Attribute new memory owner

    free(c_projections)  # Free pointer array, not actual data
    free_c_geometry(c_geometry)
    free_c_gpuids(c_gpuids)
    return projections



"""Three-dimensional Shepp-Logan head phantom
Variations of three-dimensional Shepp-Logan head phantom.

Copyright
=========
3-clause BSD License
Copyright 2021 SADAKANE, Tomoyuki
https://github.com/tsadakane/sl3d

This code is inspired by the following MATLAB codes:
 * Matthias Schabel (2021). 3D Shepp-Logan phantom (https://www.mathworks.com/matlabcentral/fileexchange/9416-3d-shepp-logan-phantom), MATLAB Central File Exchange. Retrieved April 29, 2021.
 * Patrick Bolan (2021). 3D Shepp-Logan Phantom (https://www.mathworks.com/matlabcentral/fileexchange/50974-3d-shepp-logan-phantom), MATLAB Central File Exchange. Retrieved April 29, 2021.
"""

import numpy as np

#%%
def shepp_logan_3d(size_out=128, phantom_type="yu-ye-wang", get_ellipsoids=False):
    """Three-dimensional Shepp-Logan head phantom
    Variations of three-dimensional Shepp-Logan head phantom.
    Parameters
    ==========
    phantom_type: str
        One of {"kak-slaney", "yu-ye-wang", "toft-schabel"}, optional
        Default is "yu-ye-wang"
        The type of phantom.

    size_out : int or list whose length is three, optional
        Default is [128, 128, 128]
        The number of voxels of phantom. [nVoxelZ, nVoxelY, nVoxelX]

    get_ellipsoids: bool
        Default is False.
        If False, returns the parameters of the ellipsoids as well as the phantom image.

    Returns
    =======
    If get_ellipsoids is True:
        (img_phantom, ellipsoids)
        img_phantom: numpy ndarray
        ellipsoids: list of parameters that defines ellipsoids
    else:
        img_phantom: numpy ndarray

    Notes
    =====
    For any given voxel in the output image, the voxel's value is equal to the
    sum of the additive intensity values of all ellipsoids that the voxel is a
    part of.  If a voxel is not part of any ellipsoid, its value is 0.
    The additive intensity value A for an ellipsoid can be positive or
    negative;  if it is negative, the ellipsoid will be darker than the
    surrounding voxels.
    Note that, depending on the values of A, some voxels may have values
    outside the range [0,1].
    The voxel value at (x, y, z) is img[z, y, x].
    """

    (ellipsoids, nVoxel, formula) = _parse_inputs(size_out, phantom_type)
    nVoxelX = nVoxel[2]
    nVoxelY = nVoxel[1]
    nVoxelZ = nVoxel[0]

    img_phantom = np.zeros(nVoxel, dtype=np.float32)
    range_x = np.linspace(-1, +1, nVoxelX)
    range_y = np.linspace(-1, +1, nVoxelY)
    range_z = np.linspace(-1, +1, nVoxelZ)
    mesh_z, mesh_y, mesh_x = np.meshgrid(range_z, range_y, range_x, indexing="ij")

    mesh_x = mesh_x.reshape(-1)
    mesh_y = mesh_y.reshape(-1)
    mesh_z = mesh_z.reshape(-1)

    coord = np.vstack([mesh_z, mesh_y, mesh_x])
    img_phantom = img_phantom.reshape(-1)
    for ellipsoid in ellipsoids:
        asq = ellipsoid[0] ** 2  # a^2
        bsq = ellipsoid[1] ** 2  # b^2
        csq = ellipsoid[2] ** 2  # c^2
        x0 = ellipsoid[3]  # x offset
        y0 = ellipsoid[4]  # y offset
        z0 = ellipsoid[5]  # z offset
        phi1 = ellipsoid[6] * np.pi / 180  # 1st Euler angle in radians (rotation about z-axis)
        phi2 = ellipsoid[7] * np.pi / 180  # 2nd Euler angle in radians (rotation about x'-axis)
        phi3 = ellipsoid[8] * np.pi / 180  # 3rd Euler angle in radians (rotation about z"-axis)
        A = ellipsoid[9]  # Amplitude change for this ellipsoid

        c1 = np.cos(phi1)
        s1 = np.sin(phi1)
        c2 = np.cos(phi2)
        s2 = np.sin(phi2)
        c3 = np.cos(phi3)
        s3 = np.sin(phi3)
        # Euler rotation matrix
        alpha = [  # Z      Y                    X
            [c2, -s2 * c1, s2 * s1],  # Z
            [c3 * s2, -s3 * s1 + c3 * c2 * c1, -s3 * c1 - c3 * c2 * s1],  # Y
            [s3 * s2, c3 * s1 + s3 * c2 * c1, c3 * c1 - s3 * c2 * s1],  # X
        ]
        if formula == 0:
            # Move the ellipsoid to the origin first, and rotate...
            coord_rot = np.dot(alpha, coord - np.array([[z0], [y0], [x0]]))
            idx = np.argwhere(
                (coord_rot[2, :]) ** 2 / asq
                + (coord_rot[1, :]) ** 2 / bsq
                + (coord_rot[0, :]) ** 2 / csq
                <= 1
            )
            # Naive:
            # coord_rot = np.dot(alpha, coord-np.array([[z0], [y0], [x0]])) + np.array([[z0], [y0], [x0]])
            # idx = np.argwhere((coord_rot[2,:]-x0)**2/asq + (coord_rot[1,:]-y0)**2/bsq + (coord_rot[0,:]-z0)**2/csq <= 1)
        else:
            # (x0,y0,z0) rotates too!
            coord_rot = np.dot(alpha, coord)
            idx = np.argwhere(
                (coord_rot[2, :] - x0) ** 2 / asq
                + (coord_rot[1, :] - y0) ** 2 / bsq
                + (coord_rot[0, :] - z0) ** 2 / csq
                <= 1
            )
        img_phantom[idx] += A

    img_phantom = img_phantom.reshape(nVoxel)
    if get_ellipsoids:
        return img_phantom, ellipsoids
    else:
        return img_phantom


def _parse_inputs(size_out, phantom_type):
    """
    Returns:
     tuple (ellipsoids, nVoxel)
     * ellipsoids is the m-by-10 array which defines m ellipsoids,
       where m is 10 in the cases of the variants implemented in this file.
     * nVoxel is the 3 array which defines the number of voxels
    Parameters:
     phantom_type: One of {"kak-slaney", "yu-ye-wang", "toft-schabel"}
     size_out: An int or 3-vector.
       * int : the phantom voxel will be isotropic.
       * 3-vector: the size of the phantom image [nZ, nY, nX]
    """
    if type(size_out) == int:
        nVoxel = [size_out, size_out, size_out]
    elif (type(size_out) == list or type(size_out) == tuple) and len(size_out) == 3:
        nVoxel = [size_out[0], size_out[1], size_out[2]]
    elif type(size_out) == np.array and np.size(size_out) == 3:
        nVoxel = [size_out.reshape(-1)[0], size_out.reshape(-1)[1], size_out.reshape(-1)[2]]
    else:
        nVoxel = [128, 128, 128]

    if phantom_type == "kak-slaney":
        ellipsoids = kak_slaney()
        formula = 0
    elif phantom_type == "yu-ye-wang":
        ellipsoids = yu_ye_wang()
        formula = 0
    elif phantom_type == "toft-schabel":
        ellipsoids = toft_schabel()
        formula = 1
    else:
        print(f"Unknown type {phantom_type}. yu-ye-wang is used.")
        ellipsoids = yu_ye_wang()
        formula = 0

    return (ellipsoids, nVoxel, formula)


###################################
#  Definetions of Head phantoms:  #
###################################
def kak_slaney():
    """
    The 3D Shepp-Logan head phantom. A is the relative density of water.
    Ref:
     [1] Kak AC, Slaney M, Principles of Computerized Tomographic Imaging, 1988. p.102
         http://www.slaney.org/pct/pct-errata.html
    """
    #            a        b        c      x0       y0      z0    phi1  phi2   phi3   A
    #        -------------------------------------------------------------------------
    ells = [
        [0.6900, 0.920, 0.900, 0.000, 0.000, 0.000, 0, 0, 0, 2.00],
        [0.6624, 0.874, 0.880, 0.000, 0.000, 0.000, 0, 0, 0, -0.98],
        [0.4100, 0.160, 0.210, -0.220, 0.000, -0.250, 108, 0, 0, -0.02],
        [0.3100, 0.110, 0.220, 0.220, 0.000, -0.250, 72, 0, 0, -0.02],
        [0.2100, 0.250, 0.500, 0.000, 0.350, -0.250, 0, 0, 0, 0.02],
        [0.0460, 0.046, 0.046, 0.000, 0.100, -0.250, 0, 0, 0, 0.02],
        [0.0460, 0.023, 0.020, -0.080, -0.650, -0.250, 0, 0, 0, 0.01],
        [0.0460, 0.023, 0.020, 0.060, -0.650, -0.250, 90, 0, 0, 0.01],
        [0.0560, 0.040, 0.100, 0.060, -0.105, 0.625, 90, 0, 0, 0.02],
        [0.0560, 0.056, 0.100, 0.000, 0.100, 0.625, 0, 0, 0, -0.02],
    ]
    ells = np.asarray(np.matrix(ells))
    return ells


def yu_ye_wang():
    """
    A variant of the Kak-Slaney phantom in which the contrast is improved for better
    visual perception.
    Ref:
     [2] Yu H, Ye Y, Wang G, Katsevich-Type Algorithms for Variable Radius Spiral Cone-Beam CT
         Proceedings of the SPIE, Volume 5535, p. 550-557 (2004)
    """
    ells = kak_slaney()
    ells[:, 9] = np.array([1.00, -0.80, -0.2, -0.2, 0.2, 0.2, 0.1, 0.1, 0.2, -0.2])
    # #            a      b       c      x0       y0      z0   phi1 phi2 phi3   A
    # #        -----------------------------------------------------------------
    # ells = [[ 0.6900,  0.920,  0.900,   0   ,   0    ,  0    , 0  , 0, 0,   1.0 ],
    #         [ 0.6624,  0.874,  0.880,   0   ,   0    ,  0    , 0  , 0, 0,  -0.8 ],
    #         [ 0.4100,  0.160,  0.210,  -0.22,   0    , -0.250, 108, 0, 0,  -0.2 ],
    #         [ 0.3100,  0.110,  0.220,   0.22,   0    , -0.25 , 72 , 0, 0,  -0.2 ],
    #         [ 0.2100,  0.250,  0.500,   0   ,   0.35 , -0.25 , 0  , 0, 0,   0.2 ],
    #         [ 0.0460,  0.046,  0.046,   0   ,   0.1  , -0.25 , 0  , 0, 0,   0.2 ],
    #         [ 0.0460,  0.023,  0.020,  -0.08,  -0.65 , -0.25 , 0  , 0, 0,   0.1 ],
    #         [ 0.0460,  0.023,  0.020,   0.06,  -0.65 , -0.25 , 90 , 0, 0,   0.1 ],
    #         [ 0.0560,  0.040,  0.100,   0.06,  -0.105,  0.625, 90 , 0, 0,   0.2 ],
    #         [ 0.0560,  0.056,  0.100,   0   ,   0.100,  0.625, 0  , 0, 0,  -0.2 ]]
    return ells


def toft_schabel():
    """
    The geometry of this phantom is based on the 2D phantom shown in [3] and [4].
    (Maybe this is the original Shepp-Logan head phantom?)
    In [5], the intensities of the Shepp-Logan are modified
    to yield higher contrast in the image.
    It is known as 'Modified Shepp-Logan' of the `phantom` function of "Image Processing Toolbox" for MATLAB
    In [6], it is extended to the 3D version. The parameters are as below.
    The formula of geometry transfom for this option is the same as of [6] to reproduce the result,
    while for other options, kak-slaney and yu-ye-wang, it is different.

    Ref:
     [3] Kak AC, Slaney M, Principles of Computerized Tomographic Imaging, 1988. p.55
     [4] Jain, Anil K., Fundamentals of Digital Image Processing, Englewood Cliffs, NJ, Prentice Hall, 1989, p. 439.
     [5] Toft P, The Radon Transform: Theory and Implementation, 1996.
     [6] Matthias Schabel (2021). 3D Shepp-Logan phantom
       (https://www.mathworks.com/matlabcentral/fileexchange/9416-3d-shepp-logan-phantom),
       MATLAB Central File Exchange. Retrieved April 29, 2021.
    """
    #              a     b     c     x0      y0      z0    phi1  phi2   phi3   A
    #        -----------------------------------------------------------------
    ells = [
        [0.6900, 0.9200, 0.810, 0, 0, 0, 0, 0, 0, 1.0],
        [0.6624, 0.8740, 0.780, 0, -0.0184, 0, 0, 0, 0, -0.8],
        [0.1100, 0.3100, 0.220, 0.22, 0, 0, -18, 0, 10, -0.2],
        [0.1600, 0.4100, 0.280, -0.22, 0, 0, 18, 0, 10, -0.2],
        [0.2100, 0.2500, 0.410, 0, 0.35, -0.15, 0, 0, 0, 0.1],
        [0.0460, 0.0460, 0.050, 0, 0.1, 0.25, 0, 0, 0, 0.1],
        [0.0460, 0.0460, 0.050, 0, -0.1, 0.25, 0, 0, 0, 0.1],
        [0.0460, 0.0230, 0.050, -0.08, -0.605, 0, 0, 0, 0, 0.1],
        [0.0230, 0.0230, 0.020, 0, -0.606, 0, 0, 0, 0, 0.1],
        [0.0230, 0.0460, 0.020, 0.06, -0.605, 0, 0, 0, 0, 0.1],
    ]
    return ells


if __name__ == "__main__":
    from matplotlib import pyplot as plt

    nVoxelZ = 64  # 256
    nVoxelY = 256  # 128
    nVoxelX = 256  # 64

    phantom0, e = shepp_logan_3d(
        size_out=[nVoxelZ, nVoxelY, nVoxelX], phantom_type="kak-slaney", get_ellipsoids=True
    )
    phantom1 = shepp_logan_3d(size_out=[nVoxelZ, nVoxelY, nVoxelX], phantom_type="yu-ye-wang")
    phantom2 = shepp_logan_3d(size_out=[nVoxelZ, nVoxelY, nVoxelX], phantom_type="toft-schabel")
    plt.imshow(
        np.concatenate(
            [
                phantom0[3 * phantom0.shape[0] // 8],
                phantom1[3 * phantom1.shape[0] // 8],
                phantom2[phantom2.shape[0] // 2],
            ],
            axis=1,
        )
    )
    plt.colorbar()
    plt.title("KS z=-0.25       YYW z=-0.25        TS z=0.00")
    plt.show()
    print(f"Ellipsoids of kak-slaney are\n{e}")



from __future__ import division

import matplotlib
import matplotlib.animation as animation
import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


class plotImg:  # noqa: N801
    """
    plotImg(cube, dim)
        plots figure
    default: progressive in slices following
        axis (dim)
    """

    def __init__(
        self,
        cube,
        dim=None,
        slice=None,
        step=1,
        savegif=None,
        colormap="gray",
        clims=None,
        show_plot=None,
    ):
        self.cube = cube
        self.dim = dim
        self.slice = slice
        self.dimint = None  # keeps track of what dim
        self.dimlist = ["X", "Y", "Z", "x", "y", "z", None]  # accepted parameters for dim
        self.step = step
        self.savegif = savegif
        self.colormap = colormap
        if clims is None:
            self.min_val = np.amin(self.cube)
            self.max_val = np.amax(self.cube)
        else:
            self.min_val = clims[0]
            self.max_val = clims[1]
        if show_plot is None:
            # https://matplotlib.org/stable/tutorials/introductory/usage.html#backends
            backend = matplotlib.get_backend()
            if backend in [
                "GTK3Agg",
                "GTK3Cairo",
                "MacOSX",
                "nbAgg",
                "Qt4Agg",
                "Qt4Cairo",
                "Qt5Agg",
                "Qt5Cairo",
                "TkAgg",
                "TkCairo",
                "WebAgg",
                "WX",
                "WXAgg",
                "WXCairo",
                "module://ipykernel.pylab.backend_inline",
            ]:
                self.show_plot = True
            elif backend in ["agg", "cairo", "pdf", "pgf", "ps", "svg", "template"]:
                self.show_plot = False
            else:
                self.show_plot = True
        else:
            self.show_plot = show_plot
        if self.step is None or self.step == 0:
            self.step = 1
        if self.savegif == "":
            self.savegif == None
        if self.slice is None:
            self.run()
        if self.slice is not None:
            self.slicer()

    def run(self):
        if self.dim not in self.dimlist:
            raise NameError("check inputs for dim, should be string.")

        if self.dim in [None, "X", "x"]:
            self.dimint = 2
            self.dimlist = ["->Y", "->Z", "X"]
            self.run_plot()
        if self.dim in ["Y", "y"]:
            self.dimint = 1
            self.dimlist = ["->X", "->Z", "Y"]
            self.run_plot()
        if self.dim in ["Z", "z"]:
            self.dimint = 0
            self.dimlist = ["->X", "->Y", "Z"]
            self.run_plot()

    def update_frame(self, it, fig, min_val, max_val):
        i = range(0, self.cube.shape[self.dimint])[:: self.step][it]
        fig.clf()
        axis = fig.add_subplot(1, 1, 1)
        if self.dimint == 2:
            mappable = axis.imshow(
                np.squeeze(self.cube[:, :, i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dimint == 1:
            mappable = axis.imshow(
                np.squeeze(self.cube[:, i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dimint == 0:
            mappable = axis.imshow(
                np.squeeze(self.cube[i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        axis.get_xaxis().set_ticks([])
        axis.get_yaxis().set_ticks([])
        axis.set_xlabel(self.dimlist[0])
        axis.set_ylabel(self.dimlist[1])
        axis.set_title(self.dimlist[2] + ":" + str(i))
        divider = make_axes_locatable(axis)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(mappable, cax=cax)
        # plt.pause(0.01)

    def run_plot(self):

        dim = self.cube.shape

        fig = plt.figure()
        ani = animation.FuncAnimation(
            fig,
            self.update_frame,
            fargs=(fig, self.min_val, self.max_val),
            interval=100,
            repeat_delay=1000,
            frames=len(range(0, dim[self.dimint])[:: self.step]),
        )
        if self.savegif is not None:
            ani.save(self.savegif, writer="pillow")
            self._show()
        else:
            self._show()

    def slicer(self):

        if self.dim in [None, "X", "x"]:
            plt.xlabel("Y")
            plt.ylabel("Z")
            plt.imshow(
                np.squeeze(self.cube[:, :, self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dim in ["Y", "y"]:
            plt.xlabel("X")
            plt.ylabel("Z")
            plt.imshow(
                np.squeeze(self.cube[:, self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dim in ["Z", "z"]:
            plt.xlabel("X")
            plt.ylabel("Y")
            plt.imshow(
                np.squeeze(self.cube[self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        self._show()

    def _show(self):
        if self.show_plot:
            plt.show()


plotimg = plotImg
